"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const toml_1 = __importDefault(require("toml"));
const Utils_1 = require("./Utils");
class Config {
    constructor() {
        this.defaultDataDir = Utils_1.getServiceDir('boltz-middleware');
        this.dataDir = this.defaultDataDir;
        this.init = (argv) => {
            this.parseParameters(argv);
        };
        /**
         * Parse either command line arguments or options in a config file
         */
        this.parseParameters = (parameters) => {
            // A list of all paths in which '~' should be resolved
            const pathsToResolve = [
                'logpath',
                'dbpath',
                'boltz.certpath',
                'backup.privatekeypath',
            ];
            pathsToResolve.forEach((key) => {
                if (parameters[key]) {
                    parameters[key] = Utils_1.resolveHome(parameters[key]);
                }
            });
            if (parameters.datadir) {
                this.dataDir = parameters.datadir;
                Utils_1.deepMerge(this, this.getDataDirPaths(parameters.datadir));
            }
            if (fs_1.default.existsSync(this.configpath)) {
                const config = fs_1.default.readFileSync(this.configpath, 'utf-8');
                try {
                    const parsedConfig = toml_1.default.parse(config);
                    if (!parameters.datadir && parsedConfig.datadir) {
                        Utils_1.deepMerge(this, this.getDataDirPaths(parsedConfig.datadir));
                    }
                    Utils_1.deepMerge(this, parsedConfig);
                }
                catch (error) {
                    throw `Error parsing config file in line ${error.line}:${error.column}: ${error.message}`;
                }
            }
            Utils_1.deepMerge(this, parameters);
            if (!fs_1.default.existsSync(this.dataDir)) {
                fs_1.default.mkdirSync(this.dataDir);
            }
        };
        this.getDataDirPaths = (dataDir) => {
            return {
                configpath: path_1.default.join(dataDir, 'boltz.conf'),
                logpath: path_1.default.join(dataDir, 'boltz.log'),
                dbpath: path_1.default.join(dataDir, 'boltz.db'),
                backup: {
                    privatekeypath: path_1.default.join(dataDir, 'backupPrivatekey.pem'),
                },
            };
        };
        const { configpath, logpath, dbpath, backup } = this.getDataDirPaths(this.defaultDataDir);
        this.configpath = configpath;
        this.logpath = logpath;
        this.loglevel = process.env.NODE_ENV === 'production' ? 'info' : 'debug';
        this.dbpath = dbpath;
        this.api = {
            host: '127.0.0.1',
            port: 9001,
            interval: 5,
        };
        this.boltz = {
            host: '127.0.0.1',
            port: 9000,
            certpath: path_1.default.join(Utils_1.getServiceDir('boltz'), 'tls.cert'),
        };
        this.notification = {
            token: '',
            channel: '',
            prefix: '',
            interval: 1,
        };
        this.backup = {
            email: '',
            privatekeypath: backup.privatekeypath,
            bucketname: '',
            interval: '0 0 * * *',
            backenddbpath: '',
        };
        this.currencies = [
            {
                symbol: 'BTC',
                maxSwapAmount: 100000,
                minSwapAmount: 1000,
                minWalletBalance: 1000000,
                minLocalBalance: 500000,
                minRemoteBalance: 500000,
            },
            {
                symbol: 'LTC',
                maxSwapAmount: 10000000,
                minSwapAmount: 10000,
                minWalletBalance: 100000000,
                minLocalBalance: 50000000,
                minRemoteBalance: 50000000,
            },
        ];
        this.pairs = [
            {
                base: 'LTC',
                quote: 'BTC',
                fee: 1,
            },
            {
                base: 'BTC',
                quote: 'BTC',
                fee: 0.5,
                rate: 1,
            },
            {
                base: 'LTC',
                quote: 'LTC',
                rate: 1,
                fee: 0.5,
            },
        ];
    }
}
exports.default = Config;
//# sourceMappingURL=Config.js.map