"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const bolt11_1 = __importDefault(require("@boltz/bolt11"));
const boltzrpc_pb_1 = require("./proto/boltzrpc_pb");
const idPossibilities = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
/**
 * Generate an id
 *
 * @param length how many characters the id should have
 */
exports.generateId = (length) => {
    let id = '';
    for (let i = 0; i < length; i += 1) {
        id += idPossibilities.charAt(Math.floor(Math.random() * idPossibilities.length));
    }
    return id;
};
/**
 * Get the pair id of a pair
 */
exports.getPairId = (pair) => {
    return `${pair.base}/${pair.quote}`;
};
/**
 * Get the quote and base asset of a pair id
 */
exports.splitPairId = (pairId) => {
    const split = pairId.split('/');
    return {
        base: split[0],
        quote: split[1],
    };
};
/**
 * Concat an error code and its prefix
 */
exports.concatErrorCode = (prefix, code) => {
    return `${prefix}.${code}`;
};
/**
 * Stringify any object or array
 */
exports.stringify = (object) => {
    return JSON.stringify(object, undefined, 2);
};
/**
 * Turn a map into an object
 */
exports.mapToObject = (map) => {
    const object = {};
    map.forEach((value, index) => {
        object[index] = value;
    });
    return object;
};
/**
 * Check whether a variable is a non-array object
 */
exports.isObject = (val) => {
    return (val && typeof val === 'object' && !Array.isArray(val));
};
/**
 * Recursively merge properties from different sources into a target object, overriding any
 * existing properties
 *
 * @param target The destination object to merge into.
 * @param sources The sources objects to copy from.
 */
exports.deepMerge = (target, ...sources) => {
    if (!sources.length)
        return target;
    const source = sources.shift();
    if (exports.isObject(target) && exports.isObject(source)) {
        Object.keys(source).forEach((key) => {
            if (exports.isObject(source[key])) {
                if (!target[key])
                    Object.assign(target, { [key]: {} });
                exports.deepMerge(target[key], source[key]);
            }
            else if (source[key] !== undefined) {
                Object.assign(target, { [key]: source[key] });
            }
        });
    }
    return exports.deepMerge(target, ...sources);
};
/**
 * Capitalize the first letter of a string
 */
exports.capitalizeFirstLetter = (input) => {
    return input.charAt(0).toUpperCase() + input.slice(1);
};
/**
 * Gets the home directory
 */
exports.getSystemHomeDir = () => {
    switch (os_1.default.platform()) {
        case 'win32': return process.env.LOCALAPPDATA;
        case 'darwin': return path_1.default.join(process.env.HOME, 'Library', 'Application Support');
        default: return process.env.HOME;
    }
};
/**
 * Get the data directory of a service
 */
exports.getServiceDir = (service) => {
    const homeDir = exports.getSystemHomeDir();
    const serviceDir = service.toLowerCase();
    switch (os_1.default.platform()) {
        case 'win32':
        case 'darwin':
            return path_1.default.join(homeDir, exports.capitalizeFirstLetter(serviceDir));
        default: return path_1.default.join(homeDir, `.${serviceDir}`);
    }
};
/**
 * Resolve '~' on Linux and Unix-Like systems
 */
exports.resolveHome = (filename) => {
    if (os_1.default.platform() !== 'win32') {
        if (filename.charAt(0) === '~') {
            return path_1.default.join(process.env.HOME, filename.slice(1));
        }
    }
    return filename;
};
/**
 * Convert minutes into milliseconds
 */
exports.minutesToMilliseconds = (minutes) => {
    return minutes * 60 * 1000;
};
/**
 * Convert satoshis to whole coins and remove trailing zeros
 */
exports.satoshisToCoins = (satoshis) => {
    return exports.roundToDecimals(satoshis / 100000000, 8);
};
/**
 * Round a number to a specific amount of decimals
 */
exports.roundToDecimals = (number, decimals) => {
    return Number(number.toFixed(decimals));
};
/**
 * Converts a "GetBalanceResponse" into a map
 */
exports.parseBalances = (balance) => __awaiter(this, void 0, void 0, function* () {
    const balances = new Map();
    balance.balancesMap.forEach(([key, value]) => {
        balances.set(key, value);
    });
    return balances;
});
/**
 * Gets the symbol for the fee of a swap
 */
exports.getFeeSymbol = (pairId, orderSide, isReverse) => {
    const isBuy = orderSide === boltzrpc_pb_1.OrderSide.BUY;
    const { base, quote } = exports.splitPairId(pairId);
    if (isReverse) {
        return isBuy ? quote : base;
    }
    else {
        return isBuy ? base : quote;
    }
};
/**
 * Converts the reponse of the backend method "getFeeEstimation" to an object
 */
exports.feeMapToObject = (feesMap) => {
    const response = {};
    feesMap.forEach(([symbol, fee]) => {
        response[symbol] = fee;
    });
    return response;
};
exports.getSmallestDenomination = (symbol) => {
    switch (symbol) {
        case 'LTC': return 'litoshi';
        default: return 'satoshi';
    }
};
exports.getAmountOfInvoice = (invoice) => {
    return Number(bolt11_1.default.decode(invoice).millisatoshis) / 1000;
};
//# sourceMappingURL=Utils.js.map