"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Utils_1 = require("../Utils");
const Enums_1 = require("../consts/Enums");
class Controller {
    constructor(logger, service) {
        this.logger = logger;
        this.service = service;
        // A map between the ids and HTTP responses of all pending swaps
        this.pendingSwaps = new Map();
        // A map between the ids and statuses of the swaps
        this.pendingSwapInfos = new Map();
        this.init = () => __awaiter(this, void 0, void 0, function* () {
            // Get the latest status of all swaps in the database
            const [swaps, reverseSwaps] = yield Promise.all([
                this.service.swapRepository.getSwaps(),
                this.service.reverseSwapRepository.getReverseSwaps(),
            ]);
            swaps.forEach((swap) => {
                if (swap.status) {
                    this.pendingSwapInfos.set(swap.id, { event: swap.status });
                }
            });
            reverseSwaps.forEach((reverseSwap) => {
                if (reverseSwap.status) {
                    const event = reverseSwap.status;
                    if (event !== Enums_1.SwapUpdateEvent.InvoiceSettled) {
                        this.pendingSwapInfos.set(reverseSwap.id, { event });
                    }
                    else {
                        this.pendingSwapInfos.set(reverseSwap.id, { event, preimage: reverseSwap.preimage });
                    }
                }
            });
        });
        // GET requests
        this.getPairs = (_req, res) => {
            this.successResponse(res, this.service.getPairs());
        };
        this.getFeeEstimation = (_req, res) => __awaiter(this, void 0, void 0, function* () {
            this.successResponse(res, yield this.service.getFeeEstimation());
        });
        // POST requests
        this.swapStatus = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                const { id } = this.validateBody(req.body, [
                    { name: 'id', type: 'string' },
                ]);
                const response = this.pendingSwapInfos.get(id);
                if (response) {
                    this.successResponse(res, response);
                }
                else {
                    this.errorResponse(res, `could not find swap with id: ${id}`, 404);
                }
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        });
        this.getTransaction = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                const { currency, transactionHash } = this.validateBody(req.body, [
                    { name: 'currency', type: 'string' },
                    { name: 'transactionHash', type: 'string' },
                ]);
                const response = yield this.service.getTransaction(currency, transactionHash);
                this.successResponse(res, response);
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        });
        this.broadcastTransaction = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                const { currency, transactionHex } = this.validateBody(req.body, [
                    { name: 'currency', type: 'string' },
                    { name: 'transactionHex', type: 'string' },
                ]);
                const response = yield this.service.broadcastTransaction(currency, transactionHex);
                this.successResponse(res, response);
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        });
        this.createSwap = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                const { pairId, orderSide, invoice, refundPublicKey } = this.validateBody(req.body, [
                    { name: 'pairId', type: 'string' },
                    { name: 'orderSide', type: 'string' },
                    { name: 'invoice', type: 'string' },
                    { name: 'refundPublicKey', type: 'string' },
                ]);
                const response = yield this.service.createSwap(pairId, orderSide, invoice, refundPublicKey);
                this.logger.verbose(`Created new swap with id: ${response.id}`);
                this.logger.silly(`Swap ${response.id}: ${Utils_1.stringify(response)}`);
                this.createdResponse(res, response);
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        });
        this.createReverseSwap = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                const { pairId, orderSide, claimPublicKey, amount } = this.validateBody(req.body, [
                    { name: 'pairId', type: 'string' },
                    { name: 'orderSide', type: 'string' },
                    { name: 'claimPublicKey', type: 'string' },
                    { name: 'amount', type: 'number' },
                ]);
                const response = yield this.service.createReverseSwap(pairId, orderSide, claimPublicKey, amount);
                this.logger.verbose(`Created reverse swap with id: ${response.id}`);
                this.logger.silly(`Reverse swap ${response.id}: ${Utils_1.stringify(response)}`);
                this.createdResponse(res, response);
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        });
        // EventSource streams
        this.streamSwapStatus = (req, res) => {
            try {
                const { id } = this.validateBody(req.query, [
                    { name: 'id', type: 'string' },
                ]);
                res.writeHead(200, {
                    'X-Accel-Buffering': 'no',
                    'Cache-Control': 'no-cache',
                    'Content-Type': 'text/event-stream',
                    Connection: 'keep-alive',
                });
                res.setTimeout(0);
                this.pendingSwaps.set(id, res);
                res.on('close', () => {
                    this.pendingSwaps.delete(id);
                });
            }
            catch (error) {
                this.errorResponse(res, error);
            }
        };
        /**
         * Validates that all required arguments were provided in the body correctly
         *
         * @returns the validated arguments
         */
        this.validateBody = (body, argsToCheck) => {
            const response = {};
            argsToCheck.forEach((arg) => {
                const value = body[arg.name];
                if (value !== undefined) {
                    if (typeof value === arg.type) {
                        response[arg.name] = value;
                    }
                    else {
                        throw `invalid parameter: ${arg.name}`;
                    }
                }
                else {
                    throw `undefined parameter: ${arg.name}`;
                }
            });
            return response;
        };
        this.errorResponse = (res, error, statusCode = 400) => {
            if (typeof error === 'string') {
                this.writeErrorResponse(res, statusCode, error);
            }
            else {
                if (error.details) {
                    this.writeErrorResponse(res, statusCode, error.details);
                }
                else {
                    this.writeErrorResponse(res, statusCode, error.message);
                }
            }
        };
        this.successResponse = (res, data) => {
            this.setContentTypeJson(res);
            res.status(200).json(data);
        };
        this.createdResponse = (res, data) => {
            this.setContentTypeJson(res);
            res.status(201).json(data);
        };
        this.writeErrorResponse = (res, statusCode, error) => {
            this.logger.warn(`Request failed: ${error}`);
            this.setContentTypeJson(res);
            res.status(statusCode).json({ error });
        };
        this.setContentTypeJson = (res) => {
            res.set('Content-Type', 'application/json');
        };
        this.service.on('swap.update', (id, message) => {
            this.pendingSwapInfos.set(id, message);
            const response = this.pendingSwaps.get(id);
            if (response) {
                this.logger.debug(`Swap ${id} update: ${Utils_1.stringify(message)}`);
                response.write(`data: ${JSON.stringify(message)}\n\n`);
            }
        });
    }
}
exports.default = Controller;
//# sourceMappingURL=Controller.js.map