"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const node_schedule_1 = require("node-schedule");
const storage_1 = require("@google-cloud/storage");
class BackupScheduler {
    constructor(logger, dbpath, config, boltz, report) {
        this.logger = logger;
        this.dbpath = dbpath;
        this.config = config;
        this.boltz = boltz;
        this.report = report;
        this.uploadDatabases = (date) => __awaiter(this, void 0, void 0, function* () {
            if (!this.bucket) {
                throw 'Backups are disabled because of incomplete configuration';
            }
            const dateString = BackupScheduler.getDate(date);
            this.logger.silly(`Backing up databases at: ${dateString}`);
            yield this.uploadFile(this.dbpath, dateString, true);
            if (this.config.backenddbpath !== '') {
                yield this.uploadFile(this.config.backenddbpath, dateString, false);
            }
        });
        this.uploadReport = () => __awaiter(this, void 0, void 0, function* () {
            if (!this.bucket) {
                return;
            }
            const data = yield this.report.generate();
            yield this.uploadString('report.csv', data);
        });
        this.uploadFile = (fileName, date, isMiddleware) => __awaiter(this, void 0, void 0, function* () {
            try {
                const destination = `${isMiddleware ? 'middleware' : 'backend'}/database-${date}.db`;
                yield this.bucket.upload(fileName, {
                    destination,
                });
                this.logger.silly(`Uploaded file ${fileName} to: ${destination}`);
            }
            catch (error) {
                this.logger.warn(`Could not upload file: ${error}`);
                throw error;
            }
        });
        this.uploadString = (fileName, data) => __awaiter(this, void 0, void 0, function* () {
            try {
                const file = this.bucket.file(fileName);
                yield file.save(data);
                this.logger.silly(`Uploaded data into file: ${fileName}`);
            }
            catch (error) {
                this.logger.warn(`Could not upload data to file: ${error}`);
                throw error;
            }
        });
        this.subscribeChannelBackups = () => {
            this.boltz.on('channel.backup', (currency, channelBackup, date) => __awaiter(this, void 0, void 0, function* () {
                const dateString = BackupScheduler.getDate(date || new Date());
                yield this.uploadString(`lnd/${currency}/multiChannelBackup-${dateString}`, channelBackup);
            }));
        };
        if (config.email === '' ||
            config.privatekeypath === '' ||
            config.bucketname === '') {
            logger.warn('Disabled backups because of incomplete configuration');
            return;
        }
        const storage = new storage_1.Storage({
            credentials: {
                client_email: config.email,
                private_key: fs_1.readFileSync(config.privatekeypath, 'utf-8'),
            },
        });
        this.bucket = storage.bucket(config.bucketname);
        this.subscribeChannelBackups();
        this.logger.info('Started channel backup subscription');
        this.logger.verbose(`Scheduling database backups: ${this.config.interval}`);
        node_schedule_1.scheduleJob(this.config.interval, (date) => __awaiter(this, void 0, void 0, function* () {
            yield this.uploadDatabases(date);
            yield this.uploadReport();
        }));
    }
}
BackupScheduler.getDate = (date) => {
    return `${date.getFullYear()}${BackupScheduler.addLeadingZeros(date.getMonth())}${BackupScheduler.addLeadingZeros(date.getDate())}` +
        `-${BackupScheduler.addLeadingZeros(date.getHours())}${BackupScheduler.addLeadingZeros(date.getMinutes())}`;
};
/**
 * Adds a leading 0 to the provided number if it is smalled than 10
 */
BackupScheduler.addLeadingZeros = (number) => {
    return `${number}`.padStart(2, '0');
};
exports.default = BackupScheduler;
//# sourceMappingURL=BackupScheduler.js.map