"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const grpc_1 = __importDefault(require("grpc"));
const Errors_1 = __importDefault(require("./Errors"));
const Utils_1 = require("../Utils");
const BaseClient_1 = __importDefault(require("../BaseClient"));
const Enums_1 = require("../consts/Enums");
const boltzrpc = __importStar(require("../proto/boltzrpc_pb"));
const boltzrpc_grpc_pb_1 = require("../proto/boltzrpc_grpc_pb");
var ConnectionStatus;
(function (ConnectionStatus) {
    ConnectionStatus[ConnectionStatus["Connected"] = 0] = "Connected";
    ConnectionStatus[ConnectionStatus["Disconnected"] = 1] = "Disconnected";
})(ConnectionStatus || (ConnectionStatus = {}));
exports.ConnectionStatus = ConnectionStatus;
class BoltzClient extends BaseClient_1.default {
    constructor(logger, config) {
        super();
        this.logger = logger;
        this.isReconnecting = false;
        /**
         * Connects to Boltz and subscribes to confirmed transactions and paid invoices afterwards
         */
        this.connect = () => __awaiter(this, void 0, void 0, function* () {
            if (!this.isConnected()) {
                this.boltz = new boltzrpc_grpc_pb_1.BoltzClient(this.uri, this.credentials);
                yield this.startReconnectTimer();
            }
        });
        /**
         * Gets general information about this Boltz instance and the nodes it is connected to
         */
        this.getInfo = () => {
            return this.unaryCall('getInfo', new boltzrpc.GetInfoRequest());
        };
        /**
         * Gets the balance for either all wallets or just a single one if specified
         */
        this.getBalance = (currency) => {
            const request = new boltzrpc.GetBalanceRequest();
            if (currency) {
                request.setCurrency(currency);
            }
            return this.unaryCall('getBalance', request);
        };
        /**
         * Gets a new address of a specified wallet. The "type" parameter is optional and defaults to "OutputType.LEGACY"
         */
        this.newAddress = (currency, outputType) => {
            const request = new boltzrpc.NewAddressRequest();
            request.setCurrency(currency);
            if (outputType) {
                request.setType(outputType);
            }
            return this.unaryCall('newAddress', request);
        };
        /**
         * Gets a hex encoded transaction from a transaction hash on the specified network
         */
        this.getTransaction = (currency, transactionHash) => {
            const request = new boltzrpc.GetTransactionRequest();
            request.setCurrency(currency);
            request.setTransactionHash(transactionHash);
            return this.unaryCall('getTransaction', request);
        };
        /**
         * Gets a fee estimation in satoshis per vbyte for either all currencies or just a single one if specified
         */
        this.getFeeEstimation = (currency, blocks) => {
            const request = new boltzrpc.GetFeeEstimationRequest();
            request.setCurrency(currency || '');
            request.setBlocks(blocks || 0);
            return this.unaryCall('getFeeEstimation', request);
        };
        /**
         * Broadcasts a hex encoded transaction on the specified network
         */
        this.broadcastTransaction = (currency, transactionHex) => {
            const request = new boltzrpc.BroadcastTransactionRequest();
            request.setCurrency(currency);
            request.setTransactionHex(transactionHex);
            return this.unaryCall('broadcastTransaction', request);
        };
        /**
         * Adds an entry to the list of addresses SubscribeTransactions listens to
         */
        this.listenOnAddress = (currency, address) => {
            const request = new boltzrpc.ListenOnAddressRequest();
            request.setCurrency(currency);
            request.setAddress(address);
            return this.unaryCall('listenOnAddress', request);
        };
        /**
         * Creates a new Swap from the chain to Lightning
         */
        this.createSwap = (baseCurrency, quoteCurrency, orderSide, rate, fee, invoice, refundPublicKey, outputType) => {
            const request = new boltzrpc.CreateSwapRequest();
            request.setFee(fee);
            request.setRate(rate);
            request.setInvoice(invoice);
            request.setOrderSide(orderSide);
            request.setTimeoutBlockNumber(10);
            request.setBaseCurrency(baseCurrency);
            request.setQuoteCurrency(quoteCurrency);
            request.setRefundPublicKey(refundPublicKey);
            if (outputType) {
                request.setOutputType(outputType);
            }
            return this.unaryCall('createSwap', request);
        };
        /**
         * Creates a new reverse Swap from Lightning to the chain
         */
        this.createReverseSwap = (baseCurrency, quoteCurrency, orderSide, rate, fee, claimPublicKey, amount) => {
            const request = new boltzrpc.CreateReverseSwapRequest();
            request.setFee(fee);
            request.setRate(rate);
            request.setAmount(amount);
            request.setOrderSide(orderSide);
            request.setTimeoutBlockNumber(10);
            request.setBaseCurrency(baseCurrency);
            request.setQuoteCurrency(quoteCurrency);
            request.setClaimPublicKey(claimPublicKey);
            return this.unaryCall('createReverseSwap', request);
        };
        /**
         * Subscribes to a stream of confirmed transactions to addresses that were specified with "ListenOnAddress"
         */
        this.subscribeTransactions = () => {
            if (this.transactionSubscription) {
                this.transactionSubscription.cancel();
            }
            this.transactionSubscription = this.boltz.subscribeTransactions(new boltzrpc.SubscribeTransactionsRequest(), this.meta)
                .on('data', (response) => {
                this.logger.debug(`Found transaction to address ${response.getOutputAddress()} with value ${response.getAmountReceived()} confirmed: ` +
                    `${response.getTransactionHash()}`);
                this.emit('transaction.confirmed', response.getTransactionHash(), response.getOutputAddress(), response.getAmountReceived());
            })
                .on('error', (error) => __awaiter(this, void 0, void 0, function* () {
                this.emit('status.updated', ConnectionStatus.Disconnected);
                this.logger.error(`Transaction subscription errored: ${Utils_1.stringify(error)}`);
                yield this.startReconnectTimer();
            }));
        };
        /**
         * Subscribes to a stream of settled invoices and those paid by Boltz
         */
        this.subscribeInvoices = () => {
            if (this.invoicesSubscription) {
                this.invoicesSubscription.cancel();
            }
            this.invoicesSubscription = this.boltz.subscribeInvoices(new boltzrpc.SubscribeInvoicesRequest(), this.meta)
                .on('data', (response) => {
                const invoice = response.getInvoice();
                switch (response.getEvent()) {
                    case boltzrpc.InvoiceEvent.PAID:
                        this.logger.debug(`Invoice paid: ${invoice}`);
                        this.emit('invoice.paid', invoice, response.getRoutingFee());
                        break;
                    case boltzrpc.InvoiceEvent.FAILED_TO_PAY:
                        this.logger.debug(`Failed to pay invoice: ${invoice}`);
                        this.emit('invoice.failedToPay', invoice);
                        break;
                    case boltzrpc.InvoiceEvent.SETTLED:
                        this.logger.debug(`Invoice settled: ${invoice}`);
                        this.emit('invoice.settled', invoice, response.getPreimage());
                        break;
                }
            })
                .on('error', (error) => __awaiter(this, void 0, void 0, function* () {
                this.emit('status.updated', ConnectionStatus.Disconnected);
                this.logger.error(`Invoice subscription errored: ${Utils_1.stringify(error)}`);
                yield this.startReconnectTimer();
            }));
        };
        /**
         * Subscribes to a stream of swap outputs that Boltz claims
         */
        this.subscribeClaims = () => {
            if (this.claimsSubscription) {
                this.claimsSubscription.cancel();
            }
            this.claimsSubscription = this.boltz.subscribeClaims(new boltzrpc.SubscribeClaimsRequest(), this.meta)
                .on('data', (response) => {
                const lockupTransactionHash = response.getLockupTransactionHash();
                this.logger.debug(`Claimed lockup transaction: ${lockupTransactionHash}`);
                this.emit('claim', lockupTransactionHash, response.getMinerFee());
            })
                .on('error', (error) => __awaiter(this, void 0, void 0, function* () {
                this.emit('status.updated', ConnectionStatus.Disconnected);
                this.logger.error(`Claims subscription errored: ${Utils_1.stringify(error)}`);
                yield this.startReconnectTimer();
            }));
        };
        /**
         * Subscribes to a stream of lockup transactions that Boltz refunds
         */
        this.subscribeRefunds = () => {
            if (this.refundsSubscription) {
                this.refundsSubscription.cancel();
            }
            this.refundsSubscription = this.boltz.subscribeRefunds(new boltzrpc.SubscribeRefundsRequest(), this.meta)
                .on('data', (response) => {
                const lockupTransactionHash = response.getLockupTransactionHash();
                this.logger.debug(`Refunded lockup transaction: ${lockupTransactionHash}`);
                this.emit('refund', lockupTransactionHash, response.getMinerFee());
            })
                .on('error', (error) => __awaiter(this, void 0, void 0, function* () {
                this.emit('status.updated', ConnectionStatus.Disconnected);
                this.logger.error(`Refunds subscription errored: ${Utils_1.stringify(error)}`);
                yield this.startReconnectTimer();
            }));
        };
        /**
         * Subscribes to a stream of channel backups
         */
        this.subscribeChannelBackups = () => {
            if (this.channelBackupSubscription) {
                this.channelBackupSubscription.cancel();
            }
            this.channelBackupSubscription = this.boltz.subscribeChannelBackups(new boltzrpc.SubscribeChannelBackupsRequest, this.meta)
                .on('data', (response) => {
                this.logger.debug(`New ${response.getCurrency()} channel backup`);
                this.emit('channel.backup', response.getCurrency(), response.getMultiChannelBackup());
            })
                .on('error', (error) => __awaiter(this, void 0, void 0, function* () {
                this.emit('status.updated', ConnectionStatus.Disconnected);
                this.logger.error(`Channel backup subscription errored: ${Utils_1.stringify(error)}`);
                yield this.startReconnectTimer();
            }));
        };
        this.startReconnectTimer = () => __awaiter(this, void 0, void 0, function* () {
            if (!this.isReconnecting) {
                this.isReconnecting = true;
                yield this.reconnect();
            }
        });
        this.reconnect = () => __awaiter(this, void 0, void 0, function* () {
            try {
                const getInfo = yield this.getInfo();
                this.emit('status.updated', ConnectionStatus.Connected);
                this.logger.info('Connected to Boltz');
                this.logger.verbose(`Boltz status: ${Utils_1.stringify(getInfo)}`);
                this.setClientStatus(Enums_1.ClientStatus.Connected);
                this.clearReconnectTimer();
                this.isReconnecting = false;
                this.subscribeClaims();
                this.subscribeRefunds();
                this.subscribeInvoices();
                this.subscribeTransactions();
                this.subscribeChannelBackups();
            }
            catch (error) {
                this.logger.error(`Could not connect to Boltz: ${error.message}`);
                this.logger.verbose(`Retrying in ${this.RECONNECT_INTERVAL} ms`);
                this.setClientStatus(Enums_1.ClientStatus.Disconnected);
                this.reconnectionTimer = setTimeout(this.reconnect, this.RECONNECT_INTERVAL);
            }
        });
        this.unaryCall = (methodName, params) => {
            return new Promise((resolve, reject) => {
                this.boltz[methodName](params, this.meta, (err, response) => {
                    if (err) {
                        reject(err);
                    }
                    else {
                        resolve(response.toObject());
                    }
                });
            });
        };
        const { host, port, certpath } = config;
        if (fs_1.default.existsSync(certpath)) {
            this.uri = `${host}:${port}`;
            const cert = fs_1.default.readFileSync(certpath);
            this.credentials = grpc_1.default.credentials.createSsl(cert);
            this.meta = new grpc_1.default.Metadata();
        }
        else {
            throw (Errors_1.default.COULD_NOT_FIND_FILES(certpath));
        }
    }
}
exports.default = BoltzClient;
//# sourceMappingURL=BoltzClient.js.map