"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const sequelize_1 = require("sequelize");
const Report_1 = __importDefault(require("../report/Report"));
const boltzrpc_pb_1 = require("../proto/boltzrpc_pb");
const Utils_1 = require("../Utils");
var Command;
(function (Command) {
    Command["Help"] = "help";
    // Commands that retrieve information
    Command["GetFees"] = "getfees";
    Command["SwapInfo"] = "swapinfo";
    Command["GetBalance"] = "getbalance";
    // Commands that generate a value or trigger a function
    Command["Backup"] = "backup";
    Command["NewAddress"] = "newaddress";
    Command["ToggleReverseSwaps"] = "togglereverse";
})(Command || (Command = {}));
class CommandHandler {
    constructor(logger, discord, service, boltz, backupScheduler) {
        this.logger = logger;
        this.discord = discord;
        this.service = service;
        this.boltz = boltz;
        this.backupScheduler = backupScheduler;
        /*
         * Command executors
         */
        this.help = () => __awaiter(this, void 0, void 0, function* () {
            let message = 'Commands:\n';
            this.commands.forEach((info, command) => {
                message += `\n**${command}**: ${info.description}`;
            });
            yield this.discord.sendMessage(message);
        });
        this.getBalance = () => __awaiter(this, void 0, void 0, function* () {
            const balances = yield Utils_1.parseBalances(yield this.boltz.getBalance());
            let message = 'Balances:';
            balances.forEach((balance, symbol) => {
                // tslint:disable-next-line:prefer-template
                message += `\n\n**${symbol}**\n` +
                    `Wallet: ${Utils_1.satoshisToCoins(balance.walletBalance.totalBalance)} ${symbol}`;
                if (balance.lightningBalance) {
                    const { walletBalance, channelBalance } = balance.lightningBalance;
                    const { localBalance, remoteBalance } = channelBalance;
                    // tslint:disable-next-line:prefer-template
                    message += '\n\nLND:\n' +
                        `  Wallet: ${Utils_1.satoshisToCoins(walletBalance.totalBalance)} ${symbol}\n\n` +
                        '  Channels:\n' +
                        `    Local: ${Utils_1.satoshisToCoins(localBalance)} ${symbol}\n` +
                        `    Remote: ${Utils_1.satoshisToCoins(remoteBalance)} ${symbol}`;
                }
            });
            yield this.discord.sendMessage(message);
        });
        this.getFees = () => __awaiter(this, void 0, void 0, function* () {
            let message = 'Fees:\n';
            const { swaps, reverseSwaps } = yield Report_1.default.getSuccessfulSwaps(this.service.swapRepository, this.service.reverseSwapRepository);
            const fees = this.getFeeFromSwaps(swaps, reverseSwaps);
            fees.forEach((fee, symbol) => {
                message += `\n**${symbol}**: ${Utils_1.satoshisToCoins(fee)} ${symbol}`;
            });
            yield this.discord.sendMessage(message);
        });
        this.swapInfo = (args) => __awaiter(this, void 0, void 0, function* () {
            if (args.length === 0) {
                yield this.sendCouldNotFindSwap('');
                return;
            }
            const id = args[0];
            const swap = yield this.service.swapRepository.getSwap({
                id: {
                    [sequelize_1.Op.eq]: id,
                },
            });
            if (swap) {
                yield this.sendSwapInfo(swap, false);
                return;
            }
            else {
                // Query for a reverse swap because there was no normal one found with the specified id
                const reverseSwap = yield this.service.reverseSwapRepository.getReverseSwap({
                    id: {
                        [sequelize_1.Op.eq]: id,
                    },
                });
                if (reverseSwap) {
                    yield this.sendSwapInfo(reverseSwap, true);
                    return;
                }
            }
            yield this.sendCouldNotFindSwap(id);
        });
        this.newAddress = (args) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (args.length === 0) {
                    throw 'no currency was specified';
                }
                const currency = args[0].toUpperCase();
                let outputType = boltzrpc_pb_1.OutputType.COMPATIBILITY;
                if (args.length > 1) {
                    outputType = this.getOutputType(args[1]);
                }
                const response = yield this.boltz.newAddress(currency, outputType);
                yield this.discord.sendMessage(response.address);
            }
            catch (error) {
                yield this.discord.sendMessage(`Could not generate address: ${error}`);
            }
        });
        this.toggleReverseSwaps = () => __awaiter(this, void 0, void 0, function* () {
            this.service.allowReverseSwaps = !this.service.allowReverseSwaps;
            yield this.discord.sendMessage(`${this.service.allowReverseSwaps ? 'Enabled' : 'Disabled'} reverse swaps`);
        });
        this.backup = () => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.backupScheduler.uploadDatabases(new Date());
                yield this.discord.sendMessage('Uploaded backup of databases');
            }
            catch (error) {
                yield this.discord.sendMessage(`Could not upload backup: ${error}`);
            }
        });
        /*
         * Helper functions
         */
        this.getOutputType = (type) => {
            switch (type.toLowerCase()) {
                case 'bech32': return boltzrpc_pb_1.OutputType.BECH32;
                case 'compatibility': return boltzrpc_pb_1.OutputType.COMPATIBILITY;
                case 'legacy': return boltzrpc_pb_1.OutputType.LEGACY;
            }
            throw `could not find output type: ${type}`;
        };
        this.getFeeFromSwaps = (swaps, reverseSwaps) => {
            // A map between the symbols of the currencies and the fees collected on that chain
            const fees = new Map();
            const getFeeFromSwapMap = (array, isReverse) => {
                array.forEach((swap) => {
                    const feeSymbol = Utils_1.getFeeSymbol(swap.pair, swap.orderSide, isReverse);
                    const fee = fees.get(feeSymbol);
                    if (fee) {
                        fees.set(feeSymbol, fee + swap.fee);
                    }
                    else {
                        fees.set(feeSymbol, swap.fee);
                    }
                });
            };
            getFeeFromSwapMap(swaps, false);
            getFeeFromSwapMap(reverseSwaps, true);
            return fees;
        };
        this.sendSwapInfo = (swap, isReverse) => __awaiter(this, void 0, void 0, function* () {
            yield this.discord.sendMessage(`${isReverse ? 'Reverse swap' : 'Swap'} ${swap.id}:\n\`\`\`${Utils_1.stringify(swap)}\`\`\``);
        });
        this.sendCouldNotFindSwap = (id) => __awaiter(this, void 0, void 0, function* () {
            yield this.discord.sendMessage(`Could not find swap with id: ${id}`);
        });
        this.commands = new Map([
            [Command.Help, { description: 'gets a list of all available commands', executor: this.help }],
            [Command.GetFees, { description: 'gets the accumulated fees', executor: this.getFees }],
            [Command.GetBalance, { description: 'gets the balance of the wallets and channels', executor: this.getBalance }],
            [Command.SwapInfo, { description: 'gets all available information about a (reverse) swap', executor: this.swapInfo }],
            [Command.Backup, { description: 'uploads a backup of the databases', executor: this.backup }],
            [Command.NewAddress, { description: 'generates a new address for a currency', executor: this.newAddress }],
            [Command.ToggleReverseSwaps, { description: 'enables or disables reverse swaps', executor: this.toggleReverseSwaps }],
        ]);
        this.discord.on('message', (message) => __awaiter(this, void 0, void 0, function* () {
            const args = message.split(' ');
            // Remove the first argument from the array which is the command itself
            const command = args.shift();
            if (command) {
                const commandInfo = this.commands.get(command.toLowerCase());
                if (commandInfo) {
                    this.logger.silly(`Executing command: ${command}`);
                    yield commandInfo.executor(args);
                }
            }
        }));
    }
}
exports.default = CommandHandler;
//# sourceMappingURL=CommandHandler.js.map