"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const FeeProvider_1 = __importDefault(require("./FeeProvider"));
const CryptoCompare_1 = __importDefault(require("./CryptoCompare"));
const Utils_1 = require("../Utils");
class RateProvider {
    constructor(logger, feeProvider, rateUpdateInterval, currencies) {
        this.logger = logger;
        this.feeProvider = feeProvider;
        this.rateUpdateInterval = rateUpdateInterval;
        // A map between the pair ids and the rate, limits and fees of that pair
        this.pairs = new Map();
        // A map between quote and their base assets
        // So that there is just one CryptoCompare request per quote asset needed
        this.baseAssetsMap = new Map();
        // A map of all pairs with hardcoded rates
        this.hardcodedPairs = new Map();
        // A map between assets and their limits
        this.limits = new Map();
        // A copy of the "percentageFees" Map in the FeeProvider but all values are multiplied with 100
        this.percentageFees = new Map();
        this.cryptoCompare = new CryptoCompare_1.default();
        /**
         * Gets a map of of rates from CryptoCompare for the provided pairs
         */
        this.init = (pairs) => __awaiter(this, void 0, void 0, function* () {
            this.feeProvider.percentageFees.forEach((percentage, pair) => {
                // Multiply with 100 to get the percentage
                this.percentageFees.set(pair, percentage * 100);
            });
            const minerFees = yield this.getMinerFees();
            pairs.forEach((pair) => {
                // If a pair has a hardcoded rate the CryptoCompare rate doesn't have to be queried
                if (pair.rate) {
                    this.logger.debug(`Setting hardcoded rate for pair ${pair.id}: ${pair.rate}`);
                    this.pairs.set(pair.id, {
                        rate: pair.rate,
                        limits: this.getLimits(pair.id, pair.base, pair.quote, pair.rate),
                        fees: {
                            percentage: this.percentageFees.get(pair.id),
                            minerFees: {
                                baseAsset: minerFees.get(pair.base),
                                quoteAsset: minerFees.get(pair.quote),
                            },
                        },
                    });
                    this.hardcodedPairs.set(pair.id, {
                        base: pair.base,
                        quote: pair.quote,
                    });
                    return;
                }
                const baseAssets = this.baseAssetsMap.get(pair.quote);
                if (baseAssets) {
                    baseAssets.push(pair.base);
                }
                else {
                    this.baseAssetsMap.set(pair.quote, [pair.base]);
                }
            });
            this.logger.silly(`Prepared data for requests to CryptoCompare: ${Utils_1.stringify(Utils_1.mapToObject(this.baseAssetsMap))}`);
            yield this.updateRates(minerFees);
            this.logger.debug(`Got pairs: ${Utils_1.stringify(Utils_1.mapToObject(this.pairs))}`);
            this.logger.silly(`Updating rates every ${this.rateUpdateInterval} minutes`);
            this.timer = setInterval(() => __awaiter(this, void 0, void 0, function* () {
                yield this.updateRates(yield this.getMinerFees());
            }), Utils_1.minutesToMilliseconds(this.rateUpdateInterval));
        });
        this.disconnect = () => {
            clearInterval(this.timer);
        };
        this.updateRates = (minerFees) => __awaiter(this, void 0, void 0, function* () {
            const promises = [];
            // Update the pairs with a variable rate
            this.baseAssetsMap.forEach((baseAssets, quoteAsset) => {
                promises.push(new Promise((resolve) => __awaiter(this, void 0, void 0, function* () {
                    const baseAssetsRates = yield this.cryptoCompare.getPriceMulti(baseAssets, [quoteAsset]);
                    baseAssets.forEach((baseAsset) => {
                        const pair = Utils_1.getPairId({ base: baseAsset, quote: quoteAsset });
                        const rate = baseAssetsRates[baseAsset][quoteAsset];
                        const limits = this.getLimits(pair, baseAsset, quoteAsset, rate);
                        this.pairs.set(pair, {
                            rate,
                            limits,
                            fees: {
                                percentage: this.percentageFees.get(pair),
                                minerFees: {
                                    baseAsset: minerFees.get(baseAsset),
                                    quoteAsset: minerFees.get(quoteAsset),
                                },
                            },
                        });
                    });
                    resolve();
                })));
            });
            // Update the miner fees of the pairs with a hardcoded rate
            this.hardcodedPairs.forEach(({ base, quote }, pair) => {
                const pairInfo = this.pairs.get(pair);
                pairInfo.fees.minerFees = {
                    baseAsset: minerFees.get(base),
                    quoteAsset: minerFees.get(quote),
                };
                this.pairs.set(pair, pairInfo);
            });
            yield Promise.all(promises);
            this.logger.silly('Updated pairs');
        });
        this.getLimits = (pair, base, quote, rate) => {
            const baseLimits = this.limits.get(base);
            const quoteLimits = this.limits.get(quote);
            if (baseLimits && quoteLimits) {
                // The limits we show are for the base asset and therefore to determine whether
                // the limits for the base or quote asset are higher for the minimal or lower for
                // the maximal amount we need to multiply the quote limits times (1 / rate)
                const reverseQuoteLimits = this.calculateQuoteLimits(rate, quoteLimits);
                return {
                    maximal: Math.min(baseLimits.maximal, reverseQuoteLimits.maximal),
                    minimal: Math.max(baseLimits.minimal, reverseQuoteLimits.minimal),
                };
            }
            throw `Could not get limits for pair ${pair}`;
        };
        this.calculateQuoteLimits = (rate, limits) => {
            const reverseRate = 1 / rate;
            return {
                maximal: Math.floor(reverseRate * limits.maximal),
                minimal: Math.floor(reverseRate * limits.minimal),
            };
        };
        this.parseCurrencies = (currencies) => {
            currencies.forEach((currency) => {
                this.limits.set(currency.symbol, {
                    maximal: currency.maxSwapAmount,
                    minimal: currency.minSwapAmount,
                });
            });
        };
        this.getMinerFees = () => __awaiter(this, void 0, void 0, function* () {
            const minerFees = new Map();
            for (const [symbol] of this.limits) {
                // The pair and amount can be emtpy because we just want the miner fee
                const { normal, reverseLockup } = yield this.getFeeFromProvider(symbol);
                minerFees.set(symbol, {
                    normal,
                    reverse: {
                        lockup: reverseLockup,
                        // We cannot know what kind of address the user will claim to so we just assume the worst case: P2PKH
                        //
                        // Claiming a P2WSH to a P2PKH address is about 138 bytes and to get the sats per vbyte we divide the
                        // reverse fee by the size of the reverse lockup transaction (153 vbyte)
                        claim: FeeProvider_1.default.transactionSizes.reverseClaim * (reverseLockup / FeeProvider_1.default.transactionSizes.reverseLockup),
                    },
                });
            }
            return minerFees;
        });
        this.getFeeFromProvider = (chainCurrency) => __awaiter(this, void 0, void 0, function* () {
            const [normal, reverseLockup] = yield Promise.all([
                this.feeProvider.getFee('', chainCurrency, 0, false),
                this.feeProvider.getFee('', chainCurrency, 0, true),
            ]);
            return {
                normal: normal.baseFee,
                reverseLockup: reverseLockup.baseFee,
            };
        });
        this.cryptoCompare = new CryptoCompare_1.default();
        this.parseCurrencies(currencies);
    }
}
exports.default = RateProvider;
//# sourceMappingURL=RateProvider.js.map