"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const sequelize_1 = require("sequelize");
const Logger_1 = __importDefault(require("../Logger"));
const Database_1 = __importDefault(require("../db/Database"));
const Enums_1 = require("../consts/Enums");
const SwapRepository_1 = __importDefault(require("../service/SwapRepository"));
const ReverseSwapRepository_1 = __importDefault(require("../service/ReverseSwapRepository"));
const Utils_1 = require("../Utils");
class Report {
    constructor(swapRepository, reverseSwapRepository) {
        this.swapRepository = swapRepository;
        this.reverseSwapRepository = reverseSwapRepository;
        this.generate = () => __awaiter(this, void 0, void 0, function* () {
            const { swaps: successfulSwaps, reverseSwaps: successfulReverseSwaps, } = yield Report.getSuccessfulSwaps(this.swapRepository, this.reverseSwapRepository);
            const successfulEntries = this.swapsToEntries(successfulSwaps, successfulReverseSwaps, false);
            const { swaps: failedSwaps, reverseSwaps: failedReverseSwaps, } = yield Report.getFailedSwaps(this.swapRepository, this.reverseSwapRepository);
            const failedEntries = this.swapsToEntries(failedSwaps, failedReverseSwaps, true);
            const entries = [...successfulEntries, ...failedEntries];
            entries.sort((a, b) => {
                return a.date.getTime() - b.date.getTime();
            });
            return this.arrayToCsv(entries);
        });
        this.swapsToEntries = (swaps, reverseSwaps, failed) => {
            const entries = [];
            const pushToEntries = (array, isReverse) => {
                array.forEach((swap) => {
                    const routingFee = swap['routingFee'] ? swap['routingFee'] : 0;
                    entries.push({
                        date: new Date(swap.createdAt),
                        pair: swap.pair,
                        type: this.getSwapType(swap.orderSide, isReverse),
                        orderSide: swap.orderSide === 0 ? 'buy' : 'sell',
                        // tslint:disable-next-line: object-shorthand-properties-first
                        failed,
                        minerFee: this.formatAmount(swap.minerFee ? swap.minerFee : 0),
                        routingFee: (routingFee / 1000).toFixed(3),
                        fee: this.formatAmount(swap.fee),
                        feeCurrency: Utils_1.getFeeSymbol(swap.pair, swap.orderSide, isReverse),
                    });
                });
            };
            pushToEntries(swaps, false);
            pushToEntries(reverseSwaps, true);
            return entries;
        };
        this.arrayToCsv = (entries) => {
            const lines = [];
            if (entries.length !== 0) {
                const keys = Object.keys(entries[0]);
                lines.push(keys.join(','));
            }
            entries.forEach((entry) => {
                const date = this.formatDate(entry.date);
                lines.push(`${date},${entry.pair},${entry.type},${entry.orderSide},${entry.failed},` +
                    `${entry.minerFee},${entry.routingFee},${entry.fee},${entry.feeCurrency}`);
            });
            return lines.join('\n');
        };
        this.formatDate = (date) => {
            return date.toLocaleString('en-US', { hour12: false }).replace(',', '');
        };
        this.formatAmount = (satoshis) => {
            return Utils_1.satoshisToCoins(satoshis).toFixed(8);
        };
        this.getSwapType = (orderSide, isReverse) => {
            if ((orderSide === 0 && !isReverse) || (orderSide !== 0 && isReverse)) {
                return 'Lightning/Chain';
            }
            else {
                return 'Chain/Lightning';
            }
        };
    }
}
Report.cli = (argv) => __awaiter(this, void 0, void 0, function* () {
    // Get the path to the database from the command line arguments or
    // use a default one if none was specified
    const dbPath = argv.dbpath || '~/.boltz-middleware/boltz.db';
    const db = new Database_1.default(Logger_1.default.disabledLogger, Utils_1.resolveHome(dbPath));
    yield db.init();
    const report = new Report(new SwapRepository_1.default(), new ReverseSwapRepository_1.default());
    const csv = yield report.generate();
    if (argv.reportpath) {
        fs_1.default.writeFileSync(Utils_1.resolveHome(argv.reportpath), csv);
    }
    else {
        console.log(csv);
    }
});
/**
 * Gets all successful (reverse) swaps
 */
Report.getSuccessfulSwaps = (swapRepository, reverseSwapRepository) => __awaiter(this, void 0, void 0, function* () {
    const [swaps, reverseSwaps] = yield Promise.all([
        swapRepository.getSwaps({
            status: {
                [sequelize_1.Op.eq]: Enums_1.SwapUpdateEvent.TransactionClaimed,
            },
        }),
        reverseSwapRepository.getReverseSwaps({
            status: {
                [sequelize_1.Op.eq]: Enums_1.SwapUpdateEvent.InvoiceSettled,
            },
        }),
    ]);
    return {
        swaps,
        reverseSwaps,
    };
});
/**
 * Gets all failed (reverse) swaps
 */
Report.getFailedSwaps = (swapRepository, reverseSwapRepository) => __awaiter(this, void 0, void 0, function* () {
    const [swaps, reverseSwaps] = yield Promise.all([
        swapRepository.getSwaps({
            status: {
                [sequelize_1.Op.eq]: Enums_1.SwapUpdateEvent.InvoiceFailedToPay,
            },
        }),
        reverseSwapRepository.getReverseSwaps({
            status: {
                [sequelize_1.Op.eq]: Enums_1.SwapUpdateEvent.TransactionRefunded,
            },
        }),
    ]);
    return {
        swaps,
        reverseSwaps,
    };
});
exports.default = Report;
//# sourceMappingURL=Report.js.map