"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const sequelize_1 = require("sequelize");
const events_1 = require("events");
const Errors_1 = __importDefault(require("./Errors"));
const SwapRepository_1 = __importDefault(require("./SwapRepository"));
const PairRepository_1 = __importDefault(require("./PairRepository"));
const FeeProvider_1 = __importDefault(require("../rates/FeeProvider"));
const RateProvider_1 = __importDefault(require("../rates/RateProvider"));
const PaymentRequestUtils_1 = require("./PaymentRequestUtils");
const ReverseSwapRepository_1 = __importDefault(require("./ReverseSwapRepository"));
const Enums_1 = require("../consts/Enums");
const boltzrpc_pb_1 = require("../proto/boltzrpc_pb");
const Utils_1 = require("../Utils");
class Service extends events_1.EventEmitter {
    constructor(logger, boltz, rateInterval, currencies) {
        super();
        this.logger = logger;
        this.boltz = boltz;
        this.allowReverseSwaps = true;
        this.pairs = new Map();
        this.init = (pairs) => __awaiter(this, void 0, void 0, function* () {
            // Update the pairs in the database with the ones in the config
            let dbPairs = yield this.pairRepository.getPairs();
            const isUndefinedOrNull = (value) => value === undefined || value === null;
            const comparePairArrays = (array, compare, callback) => {
                array.forEach((pair) => {
                    let inCompare = false;
                    compare.forEach((comparePair) => {
                        if (pair.base === comparePair.base &&
                            pair.quote === comparePair.quote) {
                            // If the rate is equal in config and database or not defined in the config
                            // and null in the database the database entry doesn't have to be updated
                            if (pair.rate === comparePair.rate ||
                                (isUndefinedOrNull(pair.rate)) && isUndefinedOrNull(comparePair.rate)) {
                                inCompare = true;
                            }
                        }
                    });
                    if (!inCompare) {
                        callback(pair);
                    }
                });
            };
            const promises = [];
            comparePairArrays(dbPairs, pairs, (pair) => {
                promises.push(this.pairRepository.removePair(pair));
                this.logger.debug(`Removing pair from database: ${Utils_1.stringify(pair)}`);
            });
            comparePairArrays(pairs, dbPairs, (pair) => {
                promises.push(this.pairRepository.addPair(pair));
                this.logger.debug(`Adding pair to database: ${Utils_1.stringify(pair)}`);
            });
            yield Promise.all(promises);
            if (promises.length !== 0) {
                dbPairs = yield this.pairRepository.getPairs();
            }
            this.logger.verbose('Updated pairs in database with config');
            // Make sure all pairs are supported by the backend and init the pairs array
            const { chainsMap } = yield this.boltz.getInfo();
            const chainMap = new Map();
            chainsMap.forEach(([symbol, chain]) => {
                chainMap.set(symbol, chain);
            });
            const verifyBackendSupport = (symbol) => {
                if (!chainMap.get(symbol)) {
                    throw Errors_1.default.CURRENCY_NOT_SUPPORTED_BY_BACKEND(symbol);
                }
            };
            this.feeProvider.init(pairs);
            yield this.rateProvider.init(dbPairs);
            dbPairs.forEach((pair) => {
                try {
                    verifyBackendSupport(pair.base);
                    verifyBackendSupport(pair.quote);
                    this.pairs.set(pair.id, {
                        // The values have to be set manually to avoid "TypeError: Converting circular structure to JSON" errors
                        id: pair.id,
                        base: pair.base,
                        quote: pair.quote,
                    });
                }
                catch (error) {
                    this.logger.warn(`Could not initialise pair ${pair.id}: ${error.message}`);
                }
            });
            this.logger.verbose(`Initialised ${this.pairs.size} pairs: ${Utils_1.stringify(Array.from(this.pairs.keys()))}`);
            this.listenClaims();
            this.listenRefunds();
            this.listenInvoices();
            this.listenTransactions();
        });
        /**
         * Gets all supported pairs and their conversion rates
         */
        this.getPairs = () => {
            const warnings = [];
            if (!this.allowReverseSwaps) {
                warnings.push(Enums_1.ServiceWarning.ReverseSwapsDisabled);
            }
            return {
                warnings,
                pairs: Utils_1.mapToObject(this.rateProvider.pairs),
            };
        };
        /**
         * Gets the fee estimation for all supported currencies
         */
        this.getFeeEstimation = () => __awaiter(this, void 0, void 0, function* () {
            const feeEstimation = yield this.boltz.getFeeEstimation('', 2);
            return Utils_1.feeMapToObject(feeEstimation.feesMap);
        });
        /**
         * Gets a hex encoded transaction from a transaction hash on the specified network
         */
        this.getTransaction = (currency, transactionHash) => {
            return this.boltz.getTransaction(currency, transactionHash);
        };
        /**
         * Broadcasts a hex encoded transaction on the specified network
         */
        this.broadcastTransaction = (currency, transactionHex) => __awaiter(this, void 0, void 0, function* () {
            this.logger.silly(`Broadcasting ${currency} transaction: ${transactionHex}`);
            try {
                const response = yield this.boltz.broadcastTransaction(currency, transactionHex);
                return response;
            }
            catch (error) {
                if (error.details === 'non-final (code 64)') {
                    throw Errors_1.default.TRANSACTION_NOT_FINAL();
                }
                else {
                    throw error;
                }
            }
        });
        /**
         * Creates a new Swap from the chain to Lightning
         */
        this.createSwap = (pairId, orderSide, invoice, refundPublicKey) => __awaiter(this, void 0, void 0, function* () {
            const { base, quote, rate } = this.getPair(pairId);
            const side = this.getOrderSide(orderSide);
            const chainCurrency = side === boltzrpc_pb_1.OrderSide.BUY ? quote : base;
            const lightningCurrency = side === boltzrpc_pb_1.OrderSide.BUY ? base : quote;
            const satoshi = Utils_1.getAmountOfInvoice(invoice);
            this.verifyAmount(satoshi, pairId, side, false, rate);
            const { baseFee, percentageFee } = yield this.feeProvider.getFee(pairId, chainCurrency, satoshi, false);
            const { address, redeemScript, expectedAmount, timeoutBlockHeight, } = yield this.boltz.createSwap(base, quote, side, rate, baseFee + percentageFee, invoice, refundPublicKey, boltzrpc_pb_1.OutputType.COMPATIBILITY);
            yield this.boltz.listenOnAddress(chainCurrency, address);
            const id = Utils_1.generateId(6);
            try {
                yield this.swapRepository.addSwap({
                    id,
                    invoice,
                    pair: pairId,
                    orderSide: side,
                    fee: percentageFee,
                    lockupAddress: address,
                });
            }
            catch (error) {
                throw Errors_1.default.SWAP_WITH_INVOICE_EXISTS_ALREADY();
            }
            return {
                id,
                address,
                redeemScript,
                expectedAmount,
                timeoutBlockHeight,
                bip21: PaymentRequestUtils_1.encodeBip21(chainCurrency, address, expectedAmount, `Submarine Swap to ${lightningCurrency}`),
            };
        });
        /**
         * Creates a new reverse Swap from Lightning to the chain
         */
        this.createReverseSwap = (pairId, orderSide, claimPublicKey, amount) => __awaiter(this, void 0, void 0, function* () {
            if (!this.allowReverseSwaps) {
                throw Errors_1.default.REVERSE_SWAPS_DISABLED();
            }
            const { base, quote, rate } = this.getPair(pairId);
            const side = this.getOrderSide(orderSide);
            const chainCurrency = side === boltzrpc_pb_1.OrderSide.BUY ? base : quote;
            this.verifyAmount(amount, pairId, side, true, rate);
            const { baseFee, percentageFee } = yield this.feeProvider.getFee(pairId, chainCurrency, amount, true);
            const { invoice, minerFee, amountSent, redeemScript, lockupAddress, lockupTransaction, lockupTransactionHash, } = yield this.boltz.createReverseSwap(base, quote, side, rate, baseFee + percentageFee, claimPublicKey, amount);
            yield this.boltz.listenOnAddress(chainCurrency, lockupAddress);
            const id = Utils_1.generateId(6);
            yield this.reverseSwapRepository.addReverseSwap({
                id,
                invoice,
                minerFee,
                pair: pairId,
                orderSide: side,
                fee: percentageFee,
                onchainAmount: amountSent,
                transactionId: lockupTransactionHash,
            });
            return {
                id,
                invoice,
                redeemScript,
                lockupTransaction,
                lockupTransactionHash,
            };
        });
        /**
         * Gets the base and quote asset and the rate of a currency
         */
        this.getPair = (pairId) => {
            const { base, quote } = Utils_1.splitPairId(pairId);
            const pair = this.pairs.get(pairId);
            const pairInfo = this.rateProvider.pairs.get(pairId);
            if (!pair || !pairInfo) {
                throw Errors_1.default.PAIR_NOT_SUPPORTED(pairId);
            }
            return {
                base,
                quote,
                rate: pairInfo.rate,
                limits: pairInfo.limits,
                fees: pairInfo.fees,
            };
        };
        /**
         * Verfies that the requested amount is neither above the maximal nor beneath the minimal
         */
        this.verifyAmount = (amount, pairId, orderSide, isReverse, rate) => {
            if ((!isReverse && orderSide === boltzrpc_pb_1.OrderSide.SELL) ||
                (isReverse && orderSide === boltzrpc_pb_1.OrderSide.BUY)) {
                // tslint:disable-next-line:no-parameter-reassignment
                amount = amount * (1 / rate);
            }
            const { limits } = this.getPair(pairId);
            if (limits) {
                if (Math.floor(amount) > limits.maximal)
                    throw Errors_1.default.EXCEED_MAXIMAL_AMOUNT(amount, limits.maximal);
                if (Math.ceil(amount) < limits.minimal)
                    throw Errors_1.default.BENEATH_MINIMAL_AMOUNT(amount, limits.minimal);
            }
            else {
                throw Errors_1.default.CURRENCY_NOT_SUPPORTED_BY_BACKEND(pairId);
            }
        };
        /**
         * Gets the corresponding OrderSide enum of a string
         */
        this.getOrderSide = (orderSide) => {
            switch (orderSide.toLowerCase()) {
                case 'buy': return boltzrpc_pb_1.OrderSide.BUY;
                case 'sell': return boltzrpc_pb_1.OrderSide.SELL;
                default: throw Errors_1.default.ORDER_SIDE_NOT_SUPPORTED(orderSide);
            }
        };
        this.listenTransactions = () => {
            this.boltz.on('transaction.confirmed', (outputAddress, transactionId, amountReceived) => __awaiter(this, void 0, void 0, function* () {
                const swap = yield this.swapRepository.getSwap({
                    lockupAddress: {
                        [sequelize_1.Op.eq]: outputAddress,
                    },
                });
                if (swap) {
                    if (!swap.status) {
                        yield this.swapRepository.setLockupTransactionId(swap, transactionId, amountReceived);
                        this.emit('swap.update', swap.id, { event: Enums_1.SwapUpdateEvent.TransactionConfirmed });
                    }
                }
                const reverseSwap = yield this.reverseSwapRepository.getReverseSwap({
                    transactionId: {
                        [sequelize_1.Op.eq]: transactionId,
                    },
                });
                if (reverseSwap) {
                    if (!reverseSwap.status) {
                        yield this.updateSwapStatus(reverseSwap, Enums_1.SwapUpdateEvent.TransactionConfirmed, this.reverseSwapRepository.setReverseSwapStatus);
                    }
                }
            }));
        };
        this.listenInvoices = () => {
            this.boltz.on('invoice.paid', (invoice, routingFee) => __awaiter(this, void 0, void 0, function* () {
                let swap = yield this.swapRepository.getSwap({
                    invoice: {
                        [sequelize_1.Op.eq]: invoice,
                    },
                });
                if (swap) {
                    swap = yield this.swapRepository.setInvoicePaid(swap, routingFee);
                    this.emit('swap.update', swap.id, { event: Enums_1.SwapUpdateEvent.InvoicePaid });
                }
            }));
            this.boltz.on('invoice.failedToPay', (invoice) => __awaiter(this, void 0, void 0, function* () {
                const swap = yield this.swapRepository.getSwap({
                    invoice: {
                        [sequelize_1.Op.eq]: invoice,
                    },
                });
                if (swap) {
                    const error = Errors_1.default.INVOICE_COULD_NOT_BE_PAID();
                    yield this.updateSwapStatus(swap, Enums_1.SwapUpdateEvent.InvoiceFailedToPay, this.swapRepository.setSwapStatus);
                    this.logger.info(`Swap ${swap.id} failed: ${error.message}`);
                    this.emit('swap.failed', swap, error.message);
                }
            }));
            this.boltz.on('invoice.settled', (invoice, preimage) => __awaiter(this, void 0, void 0, function* () {
                let reverseSwap = yield this.reverseSwapRepository.getReverseSwap({
                    invoice: {
                        [sequelize_1.Op.eq]: invoice,
                    },
                });
                if (reverseSwap) {
                    reverseSwap = yield this.reverseSwapRepository.setInvoiceSettled(reverseSwap, preimage);
                    this.logger.verbose(`Reverse swap ${reverseSwap.id} succeeded`);
                    this.emit('swap.update', reverseSwap.id, { preimage, event: Enums_1.SwapUpdateEvent.InvoiceSettled });
                    this.emit('swap.successful', reverseSwap);
                }
            }));
        };
        this.listenClaims = () => {
            this.boltz.on('claim', (lockupTransactionId, minerFee) => __awaiter(this, void 0, void 0, function* () {
                const swap = yield this.swapRepository.getSwap({
                    lockupTransactionId: {
                        [sequelize_1.Op.eq]: lockupTransactionId,
                    },
                });
                if (swap) {
                    yield this.swapRepository.setMinerFee(swap, minerFee);
                    this.logger.verbose(`Swap ${swap.id} succeeded`);
                    this.emit('swap.successful', swap);
                }
            }));
        };
        this.listenRefunds = () => {
            this.boltz.on('refund', (lockupTransactionId, minerFee) => __awaiter(this, void 0, void 0, function* () {
                let reverseSwap = yield this.reverseSwapRepository.getReverseSwap({
                    transactionId: {
                        [sequelize_1.Op.eq]: lockupTransactionId,
                    },
                });
                if (reverseSwap) {
                    reverseSwap = yield this.reverseSwapRepository.setTransactionRefunded(reverseSwap, minerFee);
                    const error = Errors_1.default.ONCHAIN_HTLC_TIMED_OUT();
                    this.logger.info(`Reverse swap failed: ${error.message}`);
                    this.emit('swap.update', reverseSwap.id, { event: Enums_1.SwapUpdateEvent.TransactionRefunded });
                    this.emit('swap.failed', reverseSwap, error.message);
                }
            }));
        };
        this.updateSwapStatus = (instance, event, databaseUpdate) => __awaiter(this, void 0, void 0, function* () {
            if (instance) {
                yield databaseUpdate(instance, event);
                this.emit('swap.update', instance['id'], { event });
            }
        });
        this.pairRepository = new PairRepository_1.default();
        this.swapRepository = new SwapRepository_1.default();
        this.reverseSwapRepository = new ReverseSwapRepository_1.default();
        this.feeProvider = new FeeProvider_1.default(this.logger, this.boltz);
        this.rateProvider = new RateProvider_1.default(this.logger, this.feeProvider, rateInterval, currencies);
    }
}
exports.default = Service;
//# sourceMappingURL=Service.js.map